#include <windows.h>
#include <iostream>
#include <string>
#include <stdexcept>
#include <vector>

namespace 
{
	class FileHandle 
	{
		HANDLE m_h;
		FileHandle(const FileHandle &);
		const FileHandle & operator=(const FileHandle &);
	public:
		FileHandle(HANDLE h) : m_h(h) {}
		~FileHandle()
		{
			if (m_h != INVALID_HANDLE_VALUE) CloseHandle(m_h);
		}
		operator HANDLE() { return m_h; }
	};

	typedef struct _REPARSE_DATA_BUFFER {
		DWORD  ReparseTag;
		WORD   ReparseDataLength;
		WORD   Reserved;
		WORD   SubstituteNameOffset;
		WORD   SubstituteNameLength;
		WORD   PrintNameOffset;
		WORD   PrintNameLength;
		struct {
			BYTE   DataBuffer[1];
		} GenericReparseBuffer;
	} REPARSE_DATA_BUFFER, *PREPARSE_DATA_BUFFER;

	HANDLE OpenReparsePoint(std::string path, int Access)
	{
		HANDLE h = CreateFileA(path.c_str(), Access, FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, 
			FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, 0);
		if (GetLastError()) throw std::runtime_error(std::string("Unable to open : ") + path);
		return h;
	}
	std::string get_reparse_target(HANDLE h)
	{
		const int size = 0x3fff;
		std::vector<byte> raw_data(size);
		REPARSE_DATA_BUFFER * pdb = (PREPARSE_DATA_BUFFER)(&raw_data[0]);

		DWORD bytesReturned = 0;
		BOOL res = DeviceIoControl(h, FSCTL_GET_REPARSE_POINT, 0, 0, (void*)pdb, size, &bytesReturned, 0);

		if (!res)
		{
			DWORD err = GetLastError();
			if (err == ERROR_NOT_A_REPARSE_POINT) throw std::runtime_error("Not a reparse point");
			throw std::runtime_error("Unable to query junction");
		}

		if (pdb->ReparseTag != IO_REPARSE_TAG_MOUNT_POINT) throw std::runtime_error("Not a mount point");

		std::string nonInterpretedPathPrefix = "\\??\\";
		std::wstring wtgtpath = std::wstring((wchar_t*)pdb->GenericReparseBuffer.DataBuffer + pdb->SubstituteNameOffset, (pdb->SubstituteNameLength)/sizeof(wchar_t));
		std::vector<char> tgtvec(wtgtpath.size()+1);

		int xs = wtgtpath.size();

		if (!WideCharToMultiByte(CP_UTF8, 0, wtgtpath.c_str(), xs, &tgtvec[0], xs+1, NULL, NULL)) throw std::runtime_error("Unable to convert w2a");

		std::string tgtpath(&tgtvec[0], wtgtpath.size());
		if (tgtpath.substr(0,4) == nonInterpretedPathPrefix) tgtpath = tgtpath.substr(4);
		return tgtpath;
	}
}

std::string GetReparseTarget(std::string path)
{
	FileHandle fh(OpenReparsePoint(path, GENERIC_READ));
	return get_reparse_target(fh);
}

